/*
 * Copyright (c) 2001 University of Bologna
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill;

import anthill.util.*;

/**
 *  Interface Anthill is the interface used by Anthill clients
 *  to communicate with the system. The basic operation contained
 *  in this interface enable the user to search resources, to
 *  insert resources in the Anthill network and to add knowledge
 *  of other anthills.
 *
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public interface Anthill
extends Clearable
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Anthill management
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Return the identifier of this Anthill
   */
  public AnthillId getId();

  /**
   *  Return the Manager associated to this Anthill
   */
  public Manager getManager();
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Resource management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Accept a resource request originated by an anthill client.
   *  The resource to be searched is identified by name. The method
   *  returns immediately. When the resource is found, a Resource
   *  object will be returned to the specified resource listener.
   *
   *  @param name The identifier of the resource to be searched
   *  @param listener The listener object which will be notified
   *    when the resource has been found.
   */
  public void request(String name, QueryListener listener);
  
  /**
   *  Insert a new resource in this anthill. The semantics of
   *  this operation depends on the implementation of the 
   *  Anthill, and the properties guaranteed by it.
   * 
   *  @param resource the resource to be inserted.
   */
  public void insert(Resource resource);
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Neighbour list management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add the identifier of a new neighbour anthill to the set of 
   *  neighbours known to this <CODE>Anthill</CODE>, on behalf of 
   *  an Anthill client. 
   *
   *  @param aid neighbour identifier
   */
  public void addClientNeighbour(AnthillId aid);

  /**
   *  Add the identifier of a new neighbour anthill to the set of 
   *  neighbours known to this <CODE>Anthill</CODE>, on behalf of 
   *  an Ant. 
   *
   *  @param aid neighbour identifier
   */
  public void addAntNeighbour(AnthillId aid);

  /**
   *  Remove the identifier of a neighbour anthill from the set of 
   *  neighbours known to this Anthill. 
   *
   *  @param aid neighbour identifier
   */
  public void removeNeighbour(AnthillId aid);
  
  /**
   * Return information about the neighbours known to this Anthill
   */
  public Neighbour[] getNeighbours();

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // GeneticCode management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a new genetic code to the set of genetic codes known to this anthill.
   *  When creating a new ant, an anthill may choose from one of the
   *  available genetic codes. If an implementation of <CODE>Anthill</CODE> 
   *  does not enable the use of more than one genetic code, duplicated
   *  invocations of <CODE>addGeneticCode()</CODE> must throw an 
   *  <CODE>IllegalStateException</CODE>.
   * 
   *  @param code the genetic code to be added
   *  @throws IllegalStateException if no other genetic code can be added to
   *    the anthill
   */
  public void addGeneticCode(GeneticCode code);
  
  
} // END Anthill
