/*
 * Copyright (c) 2001 University of Bologna
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.antnet;

import anthill.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 *  An <CODE>UrlResource</CODE> is an object containing the URL from which the
 *  content of a resource may be downloaded. When a <CODE>UrlResource</CODE>
 *  is inserted in a <CODE>Storage</CODE> object, the storage may perform a 
 *  local copy of the resource by downloading the contents through the
 *  input stream obtained through the <CODE>getInputStream()</CODE> method.
 *
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class UrlResource
implements Resource, Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** String identifier of the resource */
  private String name;

  /** Url of the resource */
  private URL url;
  
  /** Length of the resource */
  private long length;

  /** Mapping between attribute names and attribute values */
  private Map map;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public UrlResource()
  {
    map = Collections.synchronizedMap(new HashMap());
  }

  /**
   *  Constructs a new <CODE>UrlResource</CODE> identified by the 
   *  specified name. 
   *
   *  @param name identifier of the resource
   *  @param url  url from which the resource may be downloaded
   *  @param length length of the resource (number of bytes)
   */
  UrlResource(String name, URL url, long length)
  {
    this.name = name;
    this.url = url;
    this.length = length;
    map = Collections.synchronizedMap(new HashMap());
  }
    

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the string identifier of the object.
   */
  public String getName()
  {
    return name;
  }
  
  
  /**
   *  Returns the value associated to the specified attribute
   *  name, or null if there is no attribute with that name.
   *
   *  @param name the name of the attribute
   */
  public String getAttribute(String name)
  {
    return (String) map.get(name);
  }

  
  /**
   *  Return an input stream to download the contents of the
   *  resource. Depending on the implementation, the contents
   *  may be read from the local storage, if the resource has
   *  been copied locally; or downloaded remotely, otherwise.
   *
   *  @return an input stream to read the contents of a resource;
   */
  public InputStream getInputStream()
    throws IOException
  {
    return url.openStream();
  }

  
  /**
   *  Return the lenght of the content of the resource, if
   *  available; otherwise, it returns -1.
   */
  public long getContentLength()
  {
    return length;
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods 
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Insert a mapping between the specified attribute
   *  name and the specified attribute value.
   *
   *  @param name the attribute name
   *  @param value the attribute value
   */
  void setAttribute(String name, String value)
  {
    map.put(name, value);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshals this <CODE>UrlResource</CODE> to the specified output stream.
   */
  public void readExternal(ObjectInput in)
    throws IOException
  {

    // Read main fields
    name = in.readUTF();
    url = new URL(in.readUTF());
    length = in.readLong();

    // Read attributes
    int size = in.readInt();
    for (int i=0; i < size; i++) 
      map.put(in.readUTF(), in.readUTF());
    
  }

    
  /**
   *  Marshals this <CODE>UrlResource</CODE> to the specified output stream.
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    
    // Write main fields
    out.writeUTF(name);
    out.writeUTF(url.toString());
    out.writeLong(length);
    
    // Write attributes
    out.writeInt(map.size());
    Iterator i = map.keySet().iterator();
    while (i.hasNext()) {
      String key = (String) i.next();
      out.writeUTF(key);
      out.writeUTF((String) map.get(key));
    }
  }

} // END UrlResource

