/*
 * Copyright (c) 2001 University of Bologna
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.genie;

import anthill.*;
import anthill.gf.*;
import anthill.util.*;
import java.util.*;


/**
 *  This is the scenario
 *  
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class ScenarioImpl
implements Scenario, QueryListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  Random random = new Random();
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Reference to the ScenarioFactoryImpl which created this object */
  private ScenarioFactoryImpl scenariof;
  
  /** Peernet on which the genetic code will be evaluated */
  private Peernet net;
  
  /** Set of resources stored in the anthill network */
  private Resource[] resources;
  
  /** GeneticCode to be evaluated */
  private GeneticCode code;
  
  /** Number of successful requests */
  private int successful;
  
  /** Total number of requests */
  private int requests;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates a new ScenarioImpl.
   *
   *  @param pnf Peernet factory to be used to generate networks 
   */
  public ScenarioImpl(ScenarioFactoryImpl scenariof)
  {
    this.scenariof = scenariof;
    init(true);
  }
  
   
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Scenario
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Set the genetic code to be evaluated. In this basic implementation,
   *  only one genetic code can be present in the Anthill network at a time.
   *  If this method is called more than once, exception IllegalStateException
   *  is thrown.
   */
  public void addGeneticCode(GeneticCode code)
  {
    if (this.code != null)
      throw new IllegalStateException("ScenarioImpl can evaluate only one genetic code");
    this.code = code;
    Anthill[] nodes = net.getNodes();
    for (int i=0; i < nodes.length; i++) {
      nodes[i].addGeneticCode(code);
    }
  }

  
  /**
   *  Evaluate the fitness of the specified genetic code(s) for
   *  this scenario.
   *
   *  The fitness is evaluated in the following way:
   *  At each time, a request 
   */
  public double evaluate()
  {
    // Obtain information about the network
    Anthill[] nodes = net.getNodes();
    Manager[] managers = new Manager[nodes.length];
    for (int i=0; i < nodes.length; i++) {
      managers[i] = nodes[i].getManager();
    }
    
    // Execute the required iterations 
    int iterations = scenariof.getIterations();
    for (int i=0; i < iterations; i++) {

      // Choose random anthill and make a request
      int nant = random.nextInt(nodes.length);  
      int nres = random.nextInt(resources.length);
      nodes[nant].request(resources[nres].getName(), this);
      requests++;
      Log.println(1000, "Searching: " + resources[nres]);
      
      // Each anthill executes the code of an ant in its queue
      for (int j=0; j < managers.length; j++)
        managers[j].execute();
    }
    Log.println(300, "Found " + successful + " requests over " + requests); 
    return (double) successful / requests;
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restore the object to its initial state. The network is re-initialized
   *  as an empty network.
   */
  public void clear()
  {
    init(false);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from RequestListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Upcall method used to notify the listener that the
   *  associated query has returned the specified set
   *  of resources.
   * 
   *  @param resources The resource found
   */
  public void returnResources(Resource[] resources)
  {
    Log.println(1000, "Found: " + resources[0]);
    successful++;
  }

  /**
   *  Upcall method used to notify the listener that no
   *  more results will be delivered.
   */
  public void complete()
  {
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Initialize the object
   *
   *  @param constructor true if the method is invoked by the
   *    constructor
   */
  private void init(boolean constructor)
  {
    // Init values
    code = null;
    requests = 0;
    successful = 0;

    // Init the peernet
    if (constructor) {
      // Create the peernet 
      net = scenariof.getPeernetFactory().getPeernet();
    } else {
      // Clear the peernet 
      net.clear();
    }
      
    
    /*  
     * If the method is invoked in the constructor, create the
     * required resources; otherwise, reuse the resources 
     * previously created.
     */
    if (constructor) {
      int size = scenariof.getNresources();
      ResourceFactory resourcef = scenariof.getResourceFactory();
      resources = new Resource[size];
      for (int i=0; i < size; i++) 
        resources[i] = resourcef.getResource();
    }
      
    /*
     * Choose random anthills and put the resource in them
     * (in both cases, as the clear operation on the anthill
     * has removed every resource from the Anthill.
     */
    Anthill[] nodes = net.getNodes();
    for (int i=0; i < resources.length; i++) {
      int pos = random.nextInt(nodes.length);  
      nodes[pos].insert(resources[i]);
    }
  
  }
  
  
} // END ScenarioImpl

