/*
 * Copyright (c) 2001 University of Bologna
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nestor;

import java.io.*;
import anthill.*;

/**
 *  Class NestorAntView is a simple implementation of interface 
 *  AntView. NestorAntView follows the following rules:
 *  <ul>
 *  <li> after having invoked the move() operation, an ant
 *    cannot invoke other AntView methods except from
 *    method move().
 *  <li> the number of move() operation is not limited
 *  <li> the exact semantics of move is delegated to
 *    the Gate implementation
 *  <ul>
 *
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class NestorAntView
implements AntView
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
 
  /** Anthill to which */
  private NestorAnthill anthill;
  
  /** The descriptor of the ant currently associated with this view */
  private AntDescriptor desc;

  /** True if the resource method move() has been invoked successfully */
  private boolean moved;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *  Instantiates an AntView and stores in it the references to
   *  which the operations are delegated.
   *  <p>
   *  The constructor has package visibility to avoid
   *  instances not builded through the appropriate factory.
   *
   *  @param storage  the resource database
   *  @param gate     the communication handler
   *  @param desc     the descriptor of the ant to which this AntView
   *                  is associated
   */
  NestorAntView(NestorAnthill anthill, AntDescriptor desc)
  {
    this.anthill = anthill;
    this.desc = desc;
    moved = false;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from AntView interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the identifier of the anthill containing the
   *  ant that uses this <CODE>AntView</CODE>.
   */
  public AnthillId getId()
  {
    return anthill.getId();
  }
  

  /**
   *  Returns the ant which is associated to this AntView. 
   */
  public Ant getAnt()
  {
    return desc.getAnt();
  }

  
  /**
   *  Performs the specified query on the local resource
   *  database and returns the resources satisfying the
   *  query. The semantics of the operations is delegated
   *  to the Storage object representing the local database.
   *
   *  @param query The query to be performed
   *  @throws IllegalStateException if the ant has already moved to
   *    another anthill
   */
  public Resource[] getResources(String query)
  {
    if (moved) 
      throw new IllegalStateException("The ant already moved to another anthill");
    return anthill.getStorage().getResources(query);
  }

  
  /**
   *  Method invoked by an ant to notify the anthill that
   *  this ant, created to satisfy the given query, has 
   *  returned the specified set of resources.
   *
   *  @param query      the performed query
   *  @param resources  the resources satisfying the query
   */
  public void returnResources(String query, Resource[] resources)
  {
    if (moved) 
      throw new IllegalStateException("The ant already moved to another anthill");
    anthill.getQuerySet().returnResources(query, resources);
  }

  
  /**
   *  Add a resource to the anthill. This method enables
   *  ans to move resources from one anthill to another. 
   *  
   *  In this implementation, the semantics of this
   *  operation is delegated to the Storage object.
   *
   *  @throws IllegalStateException if the ant has already moved to
   *    another anthill
   */
  public void addResource(Resource resource)
  {
    if (moved) 
      throw new IllegalStateException("The ant already moved to another anthill");
    anthill.getStorage().addAntResource(resource);
  }
  
  
  /**
   *  Method invoked by an ant in order to add the identifier of a new 
   *  neighbour anthill to the set of neighbours known to the
   *  current anthill.
   */
  public void addNeighbour(AnthillId id)
  {
    if (moved) 
      throw new IllegalStateException("The ant already moved to another anthill");
    anthill.addAntNeighbour(id);
  }

  
  /**
   *  Method invoked by an agent in order to obtain routing information
   *  stored in the anthill.
   *
   *  @throws IllegalStateException if the ant has already moved to
   *    another anthill
   */
  public Neighbour[] getNeighbours()
  {
    if (moved) 
      throw new IllegalStateException("The ant already moved to another anthill");
    return anthill.getNeighbours();
  }
  
  
  /**
   *  Method invoked by an ant in order to be sent to another 
   *  Anthill. In this implementation, an Ant may perform at
   *  most one move. The move operation is performed at most 
   *  once.
   *
   *  @param id the id of the destination anthill
   *  @throws IOException if the ant cannot be moved to the
   *    destination anthill for communication problems;
   */
  public void move(AnthillId id)
    throws IOException
  {
    anthill.getGate().send(id, desc.getAnt());
    /*
     *  Note that moved is updated only if we are
     *  sure that the send operation has not
     *  raised an exception.
     */
    moved = true;
  }
  
  
} // END NestorAntView

