/*
 * Copyright (c) 2001 University of Bologna
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.yessensis;

import anthill.*;
import anthill.util.*;
import java.util.*;
import java.io.*;

/**
 *
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class Yessensis
implements Ant
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random number generator */
  Random random = new Random();
  
  /** Array containing the size in bits of each parameter */
  private static int[] SIZES = new int[] { 16, 16, 16, 16, 16 };
  
  /** Array containing the names of the arguments */
  private static String[] NAMES = new String[] { 
    "Cloning", "Exploration", "Alfa", "Beta", "Gamma"
  };

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters identifiers
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Cloning probability, i.e. the prob. that an ant clones a resource when 
      moving backwords */
  private static final int CLONING = 0;
  
  /** Exploration probability, i.e. the prob. that an ant chooses a neighbour
      independently from the pheromones */
  private static final int EXPLORATION = 1;
  
  private static final int ALFA   = 2;
  private static final int BETA   = 3;
  private static final int GAMMA  = 4;
  private static final int RO     = 5;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Parameters of this ant */
  private int[] parameters;
  
  /** True if this is a forward ant */
  private boolean forward; 
  
  /** True if the ant cannot clone the resources any more */
  private boolean stopcloning;

  /** Name of the resource to be searched */
  private String query;
  
  /** TabuList, i.e. link followed by this Ant */
  private AnthillId[] tlist;
  
  /** Current position in the tabulist */
  private int index;
  
  /** Current genetic code */
  private GeneticCode code;
  
  /** Resources found (when moving backword) */
  Resource[] resources;
  
  /** Pheromone factory needed to create pheromone trails */
  PheromoneFactory pheromonef;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   * 
   * @param code genetic code of the ant.
   * @param ttl initial time-to-live for this ant.
   */
  public Yessensis(GeneticCode code, int ttl, PheromoneFactory pheromonef)
  {
    this.pheromonef = pheromonef;
    this.code = code;
    tlist = new AnthillId[ttl];
    parameters = code.get(SIZES);
    index = 0;
    forward = true;
    query = null;
    code = null;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

 
  /**
   *  Return the genetic code of this ant. This method may be invoked by 
   *  an anthill on an ant to extract its genetic code and store it for 
   *  future use.  
   */
  public GeneticCode getGeneticCode()
  {
    return new GeneticCode(code);
  }

  
  /**
   *  Set the query string to be performed. The interpretation of
   *  the query string depends on the particular implementation.
   *
   *  @param query the query to be performed
   */
  public void setQuery(String query)
  {
    this.query = query;
  }
  
  
  /**
   *  This method encapsulate the computation performed by the
   *  ant. It is invoked once by an anthill when the ant is
   *  received by another anthill, or when the ant is generated
   *  locally. The run method must be executed in a controlled
   *  environment, meaning that the ant cannot be enabled to
   *  access anthill resources besides of the methods provided
   *  by the AntView interface.
   *
   *  @param view reference to the AntView object which
   *    constitutes the controlled environment of the ant
   */
  public void run(AntView view)
  {
    if (forward) 
      runForward(view);
    else
      runBackward(view);
  }
    
  
  /**
   *  This method represents the behavior of the ant when moving
   *  forward.
   */
  private void runForward(AntView view)
  {
    /*
     *  FIXME: In the computation of the best path, we are not 
     *  considering the cost of the trail; moreover, we are not
     *  checking if we are traversing an anthill already
     *  traversed
     *
     */
    
    /*
     * First of all, save the fact that I've reached this anthill 
     * in the tabu list 
     */
    tlist[index] = view.getId();
    
    /*
     * Checks if the requested resources can be found locally;
     * note that if index==0, this means that we are still in the
     * anthill which created this ant, and thus it not necessary
     * perform this check, as we assume that the anthill performed
     * the same check before creating the ant. 
     */
    if (index > 0) {
      resources = view.getResources(query);
      if (resources != null) {
        // Resources are here; go home!
        forward = false;
        Log.println(1000, "Found: go back");
        runBackward(view);
        return;
      }
    }

    /*
     *  The resource is not here, so we must move forward.
     *  First of all, we increment index. In this way, the
     *  ant will arrive to a site with index addressing an
     *  empty slot in the tabu list.
     *  Then, we check if this ant has exceeded its time
     *  to live; if so, moving to the next anthill is useless. 
     *  Note that the ttl check should be performed by
     *  anthills; we repeat it here because "bad" anthill
     *  implementation may forgot it, and this ant has
     *  limited resources.
     */
    index++;
    if (index >= tlist.length) {
      // System.err.println("Zombie!");
      return;
    }

    /* Read routing information */
    Neighbour[] neighbours = view.getNeighbours();
    int size = neighbours.length;
    float[] values = new float[size];

    /* Perform a test with the exploration probability */
    if (randomEvent(EXPLORATION)) {
      
      /* 
       *  Exploration mode. We choose among the neighbours with
       *  uniform probability.
       */
      float step = 1 / ((float) size);
      float sum = 0;
      for (int i=0; i < size; i++) {
        sum += step;
        values[i] = sum;
      }
      
    } else {
      
      /*
       *  Pheromone mode. We choose among the neghbours with
       *  probability proportional to the pheromone intensity.
       */
      float sum = 0;
      for (int i=0; i < neighbours.length; i++) {
        Pheromone pheromone = getPheromone(neighbours[i]);
        sum += pheromone.getValue(query.hashCode());
        values[i] = sum;
      }
      
      // Normalization to 1
      for (int i=0; i < neighbours.length; i++) 
        values[i] = values[i]/sum;
    }
    
      
    // Neighbour selection
    while (size > 0) {
      float r = random.nextFloat();
      int pos = 0;
      while (pos < size && r > values[pos])
        pos++;
      try {
        view.move(neighbours[pos].getId());
        return; // We exit from this cycle with this return
      } catch (IOException e) { /* Do nothing */ }
      /*
       * Communication problems with neighbour[pos]; we remove it
       * Note that although expensive, this code section should
       * be rarely executed
       */
      size--;
      if (pos < size) {
        float diff = values[pos+1] - values[pos];
        for (int i=pos; i < size; i++) {
          neighbours[i] = neighbours[i+1];
          values[i] = values[i]-diff;
        }
      }
    }
  }
  
  
  /**
   *  This method represents the behavior of the ant when moving
   *  backward.
   */
  private void runBackward(AntView view)
  {
    // Pheromone release
    Neighbour[] neighbours = view.getNeighbours();
    float sum = 0;
    for (int i=0; i < neighbours.length; i++) {
      Pheromone pheromone = getPheromone(neighbours[i]);
      float value = pheromone.getValue(query.hashCode());
      if (neighbours[i].getId().equals(tlist[index]))
        value = value * (1 + parameters[ALFA] / ((float) (1 << SIZES[ALFA])));
      sum += value;
      pheromone.setValue(query.hashCode(), value);
    }
    
    /* Normalization to 1 */
    for (int i=0; i < neighbours.length; i++) {
      Pheromone pheromone = getPheromone(neighbours[i]);
      float value = pheromone.getValue(query.hashCode());
      pheromone.setValue(query.hashCode(), value / sum);
    }

    
    // Clones the resources, if necessary
    if (!stopcloning && randomEvent(CLONING)) {
      /*
       *  Add the resources to the system
       */
      for (int i=0; i < resources.length; i++) {
        view.addResource(resources[i]);
      }
    } else {
      /*
       * If a resource is note cloned in this step,
       * it will not be cloned in the following steps
       * to reach the source.
       */
      stopcloning = true;
    }
    
    // Ant movement
    if (index > 0) {
      while (index > 0) {
        // The ant is along the path; make a further step back
        index--;
        try {
          view.move(tlist[index]);
          // Ok; we moved!
          return;
        } catch (IOException e) {
          /* 
           *  If we can't move to the next host, we try to skip
           *  it by looping in the while cycle.
           */
        }
      }
    } else {
      // The ant is arrived to the anthill from which it started
      view.returnResources(query, resources);
    }
    
  }

  

  /**
   *  Invokes method <CODE>getPheromone()</CODE> on the specified neighbour.
   *  If the result is <CODE>null</CODE>, creates a new pheromone through
   *  the pheromone factory and associates it to the neighbour
   *
   *  @param neighbour neighbour anthill containing the pheromone
   */
  private Pheromone getPheromone(Neighbour neighbour)
  {
    Pheromone pheromone = (Pheromone) neighbour.getPheromone();
    if (pheromone == null) {
      /* First time that a Yessensis visits this site */
      pheromone = pheromonef.getPheromone();
      neighbour.setPheromone(pheromone);
    }
    return pheromone;
  }

  /**
   *  Generates a random event depending on the probability
   *  enclosed into the parameter identified by <CODE>param</CODE>.
   */
  private boolean randomEvent(int param)
  {
    int limit = parameters[param];
    int value = random.nextInt(1 << SIZES[param]);
    return (value <= limit);
  }
  

  /**
   *  toString() method to obtain a string containing the parameters 
   *  of this ant
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[");
    buf.append(this.getClass().getName());
    buf.append(" - ");
    for (int i=0; i < parameters.length; i++) {
      buf.append(NAMES[i]);
      buf.append(" = ");
      buf.append(parameters[i] / ((float) (1 << SIZES[i])));
      if (i < parameters.length-1)
        buf.append(", ");
    }
    buf.append("]");
    return buf.toString();
  }
  
  
} // END Yessensis

